package Logoved::DB;

use strict;
use warnings;
use Cwd 'abs_path';
use File::Find;
use File::Basename;
use autodie qw/open close/;
use Logoved::DB::Rule;
use Logoved::DB::Rule::Type;
use Logoved::DB::Ruleset;
use parent 'Source::Shared::CLI';

our $VERSION=0.004;

my $opt_dbpath;
my @opt_define;
my @opt_addon;
my $dbpath;

our @LONGOPT=(
    'db|dbpath=s' => \$opt_dbpath,
    'D|define=s' => \@opt_define,
    'a|addon=s' => \@opt_addon,
    );

sub print_longopt {
    print '  Logoved DB options
	--db,--dbpath <path> path to logoved db
	--addon <name> load database addon <name>
';
# TODO
#	--D,--define <const> fix modifier

}

my $allrulesref;
my $errorrulesref;

my %KNOWN_KEY=map {$_=>1} qw/
DESCRIPTION
FIX
FIX_IF_FIXABLE
FORCE_STATUS
MULTILINE_PATTERN_BEGIN
MULTILINE_PATTERN_BODY
MULTILINE_PATTERN_END
PATTERN
PATTERN_SEQ
RESOLVER
SCOPE
STATISTICS
STATISTICS_HEADER
TYPE
USE_FIRST_MATCH
/;
my %_IS_BOOL_ATTR=map {$_=>1} qw/
use_first_match
    /;
my %_IS_PATTERN_ATTR=map {$_=>1} qw/
multiline_pattern_begin
multiline_pattern_body
multiline_pattern_end
pattern
pattern_seq
    /;

my %_IS_ARRAY_ATTR=map {$_=>1} qw/
description
pattern
pattern_seq
multiline_pattern_body
scope
statistics_header
    /;

my %_BOOL_MAP=((map{$_=>0} '', qw/0 OFF FALSE/),(map{$_=>1} qw/1 ON TRUE/));

sub get_fail_ruleset {
    $allrulesref=&__load_ruleset() unless $allrulesref;
    $errorrulesref=&__filter_type($allrulesref, sub {not $_[0]->is_warning_type}) unless $errorrulesref;
    return Logoved::DB::Ruleset->new(rules => $errorrulesref);
}

sub get_all_ruleset {
    $allrulesref=&__load_ruleset() unless $allrulesref;
    return Logoved::DB::Ruleset->new(rules => $allrulesref);
}

sub __filter_type {
    my ($rulesref, $type_filter_sub) = @_;
    my @filtered;
    foreach my $rule (@$rulesref) {
	if (&$type_filter_sub($rule)) {
	    push @filtered, $rule;
	}
    }
    return \@filtered;
}

sub __get_dbpath {
    return $dbpath if defined $dbpath;
    my $db_path_candidate;
    if ($opt_dbpath) {
	if (-f $opt_dbpath.'/00default') {
	    $db_path_candidate=$opt_dbpath;
	} else {
	    die "logoved db not found at $opt_dbpath\n";
	}
    } else {
	($db_path_candidate)=grep { -f $_.'/00default'} './db', '../logoved.git/db',
	    $ENV{'HOME'}.'/src/repo/logoved.git/db', '/usr/share/logoved/db';
	die "Logoved DB not found in ./db or ../logoved.git/db.
Please, clone https://git.altlinux.org/people/viy/packages/logoved.git
and add option --db /path/to/logoved.git/db to help locate Logoved DB.
" unless defined ($db_path_candidate);
    }
    $dbpath=abs_path($db_path_candidate=~s!/db$!!r);
    return $dbpath;
}

sub __load_ruleset {
    my $db_path=&__get_dbpath();
    my $templates_ref={};
    &__load_templates($db_path.'/db',$templates_ref);
    my @addons=map {split (/,/,$_)} @opt_addon;
    foreach my $addon (@addons) {
	my $templates_ref=&__load_templates(&__get_addon_path($db_path, $addon),$templates_ref);
    }
    my $rules_ref=[];
    &__load_rules($db_path.'/db',$rules_ref,$templates_ref);
    foreach my $addon (@addons) {
	&__load_rules(&__get_addon_path($db_path, $addon),$rules_ref,$templates_ref);
    }
    return $rules_ref;
}

sub __get_addon_path {
    my ($db_path,$addon)=@_;
    my $addon_path=$addon;
    $addon_path=$db_path.'/addon/'.$addon if not -d $addon;
    die "Logoved::DB: addon $addon_path not found" if not -d $addon_path;
    return $addon_path;
}

sub __load_templates {
    my ($dbpath,$templates_ref)=@_;
    find(sub {
	return if $_ ne '00default';
	my $dbfile=$File::Find::name;
	my $subpath= $dbfile=~s!^\Q$dbpath\E!!r;
	$subpath=~s!/00default$!!;
	$subpath=~s!^/!!;
	my @args;
	my $template=$templates_ref->{$subpath};
	if (!$template) {
	    $templates_ref->{$subpath}=\@args;
	} else {
	    push @$template, @args;
	}
	&__load_logoved_file($dbfile, \@args);
	 }, $dbpath);
}

sub __load_rules {
    my ($dbpath, $rules_ref, $templates_ref)=@_;
    my $hookpath;
    find(sub {
	return if $_ !~ m!\.logoved$!;
	my $dbfile=$File::Find::name;
	my $subpath=$dbfile=~s!^\Q$dbpath\E/!!r;
	$subpath=~s!\.logoved$!!;
	my $entryname= $subpath=~s!/!::!gr;
	my @args=(name => $entryname,
		  subpath => $subpath,
	    );
	&__load_logoved_file($dbfile,\@args);
	my @mix;
	my @subpath_components=split('/',$subpath);
	my $subpath_prefix='';
	LOOP: while (1) {
	    my $template=$templates_ref->{$subpath_prefix};
	    push @mix, @$template if $template;
	    if ($subpath_prefix eq '') {
		$subpath_prefix=shift @subpath_components;
	    } elsif (0==@subpath_components) {
		last LOOP;
	    } else {
		$subpath_prefix .= '/'.(shift @subpath_components);
	    }
	}
	push @mix, @args;
	my $rule=Logoved::DB::Rule->new(@mix);
	my $type=$rule->type;
	die 'Oops! no type in rule: ',$rule->subpath,'.logoved' if not $type;
	die 'Oops! BAD type ',$rule,' in rule: ',$rule->subpath,'.logoved' unless Logoved::DB::Rule::Type::is_type($type);
	my $pattern_seq=$rule->pattern_seq;
	die 'Oops! single PATTERN_SEQ: ',$rule->subpath,'.logoved' if $pattern_seq and @$pattern_seq<2;
	my $fix=$rule->fix;
	die 'Oops! FIX but not error: ',$rule->subpath,'.logoved' if $fix and not Logoved::DB::Rule::Type::is_error($type);
	if ($fix) {
	    if ($fix =~ /^(?:log)?hook\s+(\S+)/) {
		my $hook=$1;
		unless ($hookpath) {
		    # lazy calculation
		    $hookpath='/../hooks/';
		    $hookpath='/../../hooks/' if $dbpath=~m!\baddon/!;
		}
		warn "WARNING: hook $dbpath$hookpath$hook not found\n" if ! -e $dbpath.$hookpath.$hook;
	    } elsif ($fix =~ /^sourcedep\s+/) {
		my $resolver=$fix=~s/^sourcedep/sourcedep-resolve/r;
		$rule->fix_if_resolved(1);
		$rule->resolver($resolver);
	    }
	}
	#use Data::Dumper;
	#print Dumper($rule) if $rule->subpath =~/cpu-time/;
	push @$rules_ref, $rule;
	 }, $dbpath);
}

sub __load_logoved_file {
    my ($dbfile,$argsref)=@_;
    open (my $fn, '<', $dbfile);
    local $_;
    my %collector;
    my %checker;
    while (<$fn>) {
	chomp;
	next if /^\s*(?:#|$)/;
	s/\s*$//;
	last if /^__END__/;
	if (my ($bigname, $value) = /^(\w+)=(.*)$/) {
	    warn "WARNING: unknown key $bigname in $dbfile\n" if ! $KNOWN_KEY{$bigname};
	    my $name=lc($bigname);
	    $value=qr/$value/ if $_IS_PATTERN_ATTR{$name};
	    if ($_IS_ARRAY_ATTR{$name}) {
		push @{$collector{$name}}, $value;
	    } elsif ($_IS_BOOL_ATTR{$name}) {
		warn 'WARNING: value should be 0 or 1 for key ',$name,' in ',$dbfile if $value ne '0' and $value ne '1';
		die 'ERROR: unrecognized bool value ',$value,' for key ',$name,' in ',$dbfile if not exists($_BOOL_MAP{$value});
		$collector{$name}=$_BOOL_MAP{$value};
	    } else {
		warn "WARNING: key $bigname in $dbfile: duplicate is ignored\n" if $checker{$name};
		push @$argsref, $name, $value;
		$checker{$name}=1;
	    }
	} else {
	    die "Oops: syntax error in $dbfile: $_";
	}
    }
    close($fn);
    foreach my $key (keys(%collector)) {
	push @$argsref, $key => $collector{$key};
    }
}

1;
